import minmlst.config as c
from scipy.spatial.distance import pdist
from scipy.cluster.hierarchy import linkage, fcluster
from sklearn.metrics.cluster import adjusted_rand_score
import numpy as np
from numpy.random import permutation
from joblib import Parallel, delayed
import pandas as pd
import pickle
from os.path import join
from minmlst.utils import *
import matplotlib.pyplot as plt


def save_temp_files(percentiles, thresholds, z, num_of_genes):
    create_dir_if_not_exists(c.TEMP_FOLDER)
    thres_per_perc = dict(zip(percentiles, thresholds))
    with open(join(c.TEMP_FOLDER, f"thresholds_{num_of_genes}" + '.pickle'), 'wb') as handle:
        pickle.dump(thres_per_perc, handle, protocol=pickle.HIGHEST_PROTOCOL)
    with open(join(c.TEMP_FOLDER, f"z_{num_of_genes}" + '.pickle'), 'wb') as handle:
        pickle.dump(z, handle, protocol=pickle.HIGHEST_PROTOCOL)


def simulation_study_ARI(partition_A, partition_B, ARI_0, num_of_samples):
    ARI_dist = np.empty(num_of_samples, dtype=float)
    for i in range(num_of_samples):
        ARI_dist[i] = adjusted_rand_score(permutation(partition_A), permutation(partition_B))
    m = np.average(ARI_dist)
    std = np.std(ARI_dist)
    if std == 0:
        std = np.finfo(np.float32).eps
    NARI_0 = (ARI_0 - m) / std
    NARI_dist = (ARI_dist - m) / std
    p_value = len(NARI_dist[NARI_dist > NARI_0]) / len(NARI_dist)

    return p_value


def is_next_better(best_ARI_vec, next_ARI_vec):
    minimal_diff = 0
    diff = sum(next_ARI_vec - best_ARI_vec)
    return diff >= minimal_diff


def calc_ARI_pv(res, CT, predicted_CT_lst, percentiles, thresholds, simulated_samples):
    for idx, predicted_CT in enumerate(predicted_CT_lst):
        ARI = adjusted_rand_score(CT, predicted_CT)
        res.update({f"ARI_perc_{format(percentiles[idx], '.15g')}": ARI})
        res.update({f"threshold_value_perc_{format(percentiles[idx], '.15g')}": thresholds[idx]})
        if simulated_samples > 0:
            p_value = simulation_study_ARI(CT, predicted_CT, ARI, simulated_samples)
            res.update({f"pv_perc_{format(percentiles[idx], '.15g')}": p_value})


def calc_ARI_pv_vec(num_of_genes, perc, CT, simulated_samples):
    res = {'num_of_genes': num_of_genes}
    with open(join(c.TEMP_FOLDER, f"thresholds_{num_of_genes}" + '.pickle'), 'rb') as handle:
        thres_per_perc = pickle.load(handle)
    with open(join(c.TEMP_FOLDER, f"z_{num_of_genes}" + '.pickle'), 'rb') as handle:
        z = pickle.load(handle)
    curr_thres = thres_per_perc[perc]
    predicted_CT_lst = [fcluster(Z=z, t=curr_thres, criterion='distance')]
    calc_ARI_pv(res, CT, predicted_CT_lst, [perc], [curr_thres], simulated_samples)

    return res


def parse_ans(ans, perc, simulated_samples):
    ans_df = pd.DataFrame(ans)
    _ans = {"perc": perc, "ARI_vec": ans_df[f"ARI_perc_{format(perc, '.15g')}"]}
    if simulated_samples > 0:
        _ans.update({"pv_vec": ans_df[f"pv_perc_{format(perc, '.15g')}"]})
    return _ans


def find_percentile(results, CT, percentiles_to_check, simulated_samples, n_jobs):
    best_perc = percentiles_to_check[0]
    next_perc = percentiles_to_check[1]
    _best = {"perc": best_perc, "ARI_vec": results[f"ARI_perc_{format(best_perc, '.15g')}"]}
    _next = {"perc": next_perc, "ARI_vec": results[f"ARI_perc_{format(next_perc, '.15g')}"]}
    if simulated_samples > 0:
        _best.update({"pv_vec": results[f"pv_perc_{format(best_perc, '.15g')}"]})
        _next.update({"pv_vec": results[f"pv_perc_{format(next_perc, '.15g')}"]})
    next_better = is_next_better(_best["ARI_vec"], _next["ARI_vec"])

    if len(percentiles_to_check) > 2:
        potential_perc = list(percentiles_to_check[2:])
        while next_better and potential_perc:
            _best = _next
            next_perc = potential_perc.pop(0)
            try:
                ans = Parallel(n_jobs=n_jobs, verbose=5, max_nbytes=None)(
                    delayed(calc_ARI_pv_vec)(num_of_genes, next_perc, CT, simulated_samples) for num_of_genes in results['num_of_genes'])
            except Exception as ex:
                print(f"Error - unable to perform parallel computing due to: {ex}")
                print(f"Running serial computation instead")
                ans = []
                for num_of_genes in results['num_of_genes']:
                    a = calc_ARI_pv_vec(num_of_genes, next_perc, CT, simulated_samples)
                    ans = ans + [a]
            _next = parse_ans(ans, next_perc, simulated_samples)
            next_better = is_next_better(_best["ARI_vec"], _next["ARI_vec"])

    if next_better:
        _best = _next

    remove_dir(c.TEMP_FOLDER)
    res = {'num_of_genes': results['num_of_genes'], f"ARI_perc_{format(_best['perc'], '.15g')}": _best["ARI_vec"]}
    if simulated_samples > 0:
        res.update({f"pv_perc_{format(_best['perc'], '.15g')}": _best["pv_vec"]})

    return pd.DataFrame(res)


def hierarchical_clustering(CT, x, num_of_genes, gene_importance, linkage_method, percentiles, find_percentile,
                            percentiles_to_check, simulated_samples):
    res = {'num_of_genes': num_of_genes}
    curr_genes = gene_importance['gene'][0:num_of_genes]
    curr_x = x.loc[:, curr_genes]
    distances = pdist(X=curr_x, metric=c.DISTANCE_METRIC)
    z = linkage(y=distances, method=linkage_method)

    if find_percentile:
        # save percentiles and thresholds as temp files
        percentiles = percentiles_to_check
        thresholds = np.percentile(a=distances, q=percentiles)
        save_temp_files(percentiles, thresholds, z, num_of_genes)
        # percentiles 0.5 and 1 are calculated as a baseline
        percentiles = percentiles[:2]
        thresholds = thresholds[:2]
    else:
        thresholds = np.percentile(a=distances, q=percentiles)
    predicted_ST_lst = [fcluster(Z=z, t=t, criterion='distance') for t in thresholds]
    calc_ARI_pv(res, CT, predicted_ST_lst, percentiles, thresholds, simulated_samples)

    return res


def reorder_analysis_res(df):
    cols = list(df.columns.values)
    cols.remove('num_of_genes')
    return df[['num_of_genes'] + cols]


def plot_res(analysis_res, measure):
    title = f'Results per number of genes (measure = {measure})'
    x_label = 'Number of genes'
    y_label = 'Adjusted Rand Index  or  p-value'

    fig, ax = plt.subplots(nrows=1, ncols=1, dpi=150)
    ax.spines["top"].set_visible(False)
    ax.spines["right"].set_visible(False)
    ax.get_xaxis().tick_bottom()
    ax.get_yaxis().tick_left()
    ax.set_title(title)
    ax.set_ylabel(y_label)
    ax.set_xlabel(x_label)
    ax.set_yticks(np.arange(0, 1.2, 0.2))
    ax.set_ylim((-0.05, 1))

    x_col = 'num_of_genes'
    y_cols = list(analysis_res.columns.values)
    y_cols.remove(x_col)
    for y_col in y_cols:
        ax.plot(analysis_res[x_col], analysis_res[y_col], label=y_col, marker='o', linestyle='--')
    ax.legend(frameon=True, bbox_to_anchor=(1, 0.5), loc="center left")
    plt.show()



import pandas as pd
import numpy as np
import collections


def validate_data(data):
    if not isinstance(data, pd.DataFrame):
        raise ValueError(f"Error: 'data' must be of type <class 'pandas.core.frame.DataFrame'>,"
                         f" got {type(data)}.")
    elif data.empty:
        raise ValueError(f"Error: 'data' is empty.")
    invalid = [not (np.issubdtype(t, np.integer)) for t in data.dtypes]
    if sum(invalid) > 0:
        raise ValueError(f"Error: 'data' contains non-integer elements. Invalid columns and types:"
                         f"\n{data.dtypes[invalid]}")
    if 0 in list(data.iloc[:, -1]):
        raise ValueError(f"Error: strain-type column (last) contains missing values, i.e value = 0")


def validate_percentiles(percentiles):
    valid = False
    if np.issubdtype(type(percentiles), np.number) and 0 < percentiles < 100:
        valid = True
    if isinstance(percentiles, (collections.Sequence, np.ndarray)) and len(percentiles) > 0:
        for p in percentiles:
            if (not np.issubdtype(type(p), np.number)) or (not(0 < p < 100)):
                valid = False
                break
            else:
                valid = True
    if not valid:
        raise ValueError(f"Error: 'percentiles' must be a number or an array of numbers, "
                         f"as each number N is 0 < N < 100.")


def validate_input_gi(data, measures, max_depth, learning_rate, stopping_method, stopping_rounds):
    print("Input validation")
    # data
    validate_data(data)
    # measures
    valid_measures = ['shap', 'weight', 'gain', 'cover', 'total_gain', 'total_cover']
    if not isinstance(measures, (collections.Sequence, np.ndarray)) or len(measures) == 0:
        raise ValueError(f"Error: 'measures' must be a non-empty array. Valid elements are: {valid_measures}.")
    for m in measures:
        if m not in valid_measures:
            raise ValueError(f"Error: 'measures' contains invalid element {m}. Valid elements are: {valid_measures}.")
    # max_depth
    if not np.issubdtype(type(max_depth), np.integer) or max_depth < 0:
        raise ValueError(f"Error: 'max_depth' must be of type int and greater equal to 0, got {type(max_depth)}")
    # learning_rate
    if not np.issubdtype(type(learning_rate), np.number) or learning_rate < 0:
        raise ValueError(f"Error: 'learning_rate' must be a number greater equal to 0, got {type(learning_rate)}")
    # stopping_method
    if stopping_method not in ['num_boost_round', 'early_stopping_rounds']:
        raise ValueError(f"Error: 'stopping_method' must be 'num_boost_round' or 'early_stopping_rounds' (type str)")
    # stopping_rounds
    if not np.issubdtype(type(stopping_rounds), np.integer) or stopping_rounds <= 0:
        raise ValueError(f"Error: 'stopping_rounds' must be of type int and greater than 0, got {type(stopping_rounds)}")


def validate_input_gra(data, gene_importance, measure, reduction, linkage_method, percentiles, percentiles_to_check,
                       simulated_samples, n_jobs):
    print("Input validation")
    # data
    validate_data(data)
    # measure
    valid_measures = ['shap', 'weight', 'gain', 'cover', 'total_gain', 'total_cover']
    if measure not in valid_measures:
        raise ValueError(f"Error: measure must be either 'shap', 'weight', 'gain', 'cover', 'total_gain' "
                         f"or 'total_cover'.")
    # gene_importance
    if not isinstance(gene_importance, pd.DataFrame):
        raise ValueError(f"Error: 'gene_importance' must be in the format returned by 'gene_importance' function.")
    gi_cols = gene_importance.columns.values
    if len(gi_cols) < 2 or gi_cols[0] != 'gene':
        raise ValueError(f"Error: 'gene_importance' must be in the format returned by 'gene_importance' function.")
    gi_measures = [col.replace("importance_by_", "") for col in gi_cols[1:]]
    for m in gi_measures:
        if m not in valid_measures:
            raise ValueError(f"Error: 'gene_importance' must be in the format returned by 'gene_importance' function.")
    invalid = [not (np.issubdtype(t, np.number)) for t in gene_importance.dtypes[1:]]
    if sum(invalid) > 0:
        raise ValueError(f"Error: 'gene_importance' contains non-numeric importance scores. "
                         f"Invalid columns and types: \n{[False] + gene_importance.dtypes[invalid]}")
    gi_genes = set(gene_importance.iloc[:, 0])
    data_genes = set(data.columns.values[:-1])
    if len(gi_genes - data_genes) + len(data_genes - gi_genes) > 0:
        raise ValueError(f"Error: genes in 'data' and 'gene_importance' do not match.")
    # measure
    if measure not in gi_measures:
        raise ValueError(f"Error: 'measure' must be included in the 'gene_importance' results -> {gi_measures}.")
    # reduction
    if (not np.issubdtype(type(reduction), np.number)) or (reduction <= 0):
        raise ValueError(f"Error: 'reduction' must be a positive number. Use int for number of genes, or float for"
                         f" percentage of genes to be reduced.")
    # linkage_method
    valid_methods = ['single', 'complete', 'average', 'weighted', 'centroid', 'median', 'ward']
    if linkage_method not in valid_methods:
        raise ValueError(f"Error: linkage_method must be either 'single', 'complete', 'average', 'weighted', "
                         f"'centroid', 'median' or 'ward'.")
    # percentiles
    validate_percentiles(percentiles)
    # percentiles_to_check
    if not isinstance(percentiles_to_check, (collections.Sequence, np.ndarray)) or len(percentiles_to_check) < 2:
        raise ValueError(f"Error: 'percentiles_to_check' must be an array containing at least 2 numbers (percentiles), got: {percentiles_to_check}")
    for p in percentiles_to_check:
        if (not np.issubdtype(type(p), np.number)) or (not (0 < p < 100)):
            raise ValueError(f"Error: each element in 'percentiles_to_check' must be a number in the range of (0, 100), got: {p}")
    # simulated_samples
    if not np.issubdtype(type(simulated_samples), np.integer):
        raise ValueError(f"Error: 'simulation_iter' must be of type int, got {type(simulated_samples)}")
    if 0 < simulated_samples < 1000:
        print("Warning: for the significance of the p.v results, a larger number of simulated samples (1000, say) is "
              "required. \nSee- https://www.sciencedirect.com/science/article/pii/S0950329313000852\n")
    # n_jobs
    if not np.issubdtype(type(n_jobs), np.integer):
        raise ValueError(f"Error: 'n_jobs' must be of type int, got {type(n_jobs)}")
    return min(n_jobs, 60), sorted(percentiles_to_check)  # limit n_jobs to 60, sort 'percentiles_to_check'



"""
Machine learning module for finding a minimal MLST scheme for bacterial strain typing
=====================================================================================

minMLST is a machine-learning based methodology for identifying a minimal subset of genes that preserves high
discrimination among bacterial strains. It combines well known machine-learning algorithms and approaches such as
XGBoost, distance-based hierarchical clustering, and SHAP.
minMLST quantifies the importance level of each gene in an MLST scheme and allows the user to investigate the trade-off
between minimizing the number of genes in the scheme vs preserving a high resolution among strains.

"""
# Let users know if they're missing any of the hard dependencies
hard_dependencies = ("shap", "xgboost", "sklearn")
missing_dependencies = []

for dependency in hard_dependencies:
    try:
        __import__(dependency)
    except ImportError as e:
        missing_dependencies.append(dependency)

if missing_dependencies:
    raise ImportError("Missing required dependencies {0}".format(missing_dependencies))
del hard_dependencies, dependency, missing_dependencies

from minmlst.gene_importance import *
from minmlst.clustering import *
from minmlst.input_validation import *
import pandas as pd
import numpy as np
import os
import traceback
from joblib import Parallel, delayed
import multiprocessing as mp
# region set random seeds
os.environ['PYTHONHASHSEED'] = '0'
np.random.seed(c.SEED)
# endregion set random seeds


def gene_importance(data, measures, max_depth=6, learning_rate=0.3, stopping_method='num_boost_round',
                    stopping_rounds = 100):
    """
    This function provides a ranking of gene importance according to selected measures: 'shap', 'weight', 'gain',
    'cover', 'total_gain' or 'total_cover'.
        'shap' - the mean magnitude of the SHAP values, i.e. the mean absolute value of the SHAP values of a given gene
                 (See: http://papers.nips.cc/paper/7062-a-unified-approach-to-interpreting-model-predictions).
        'weight' - the number of times a given gene is used to split the data across all splits.
        'gain' (or 'total gain') - the average (or total) gain is the average (or total) reduction of Multiclass Log Loss
                                   contributed by a given gene across all splits.
        'cover' (or 'total cover') - the average (or total) quantity of observations concerned by a given gene across all splits.

    As a pre-step, CTs (cluster types) with a single representative isolate are filtered from the dataset.
    Next, an XGBoost model is trained with parameters 'max_depth', 'learning_rate', 'stopping_method' and 'stopping_rounds' -
    more information about XGBoost parameters can be found here: https://xgboost.readthedocs.io/en/latest/python/python_api.html.
    Model's performance is evaluated by Multi-class log loss over a test set.
    Finally, gene importance values are measured for the trained model and provided as a DataFrame output.

    :param data (DataFrame): DataFrame in the shape of (m,n).
                             (n-1) columns of genes, last column (n) must contain the CT (cluster type).
                             Each row (m) represents a profile of a single isolate.
                             Data types should be integers only.
                             Missing values should be represented as 0, no missing values are allowed for the CT (last
                             column).
    :param measures (array): an array containing at least one of the following measures (str type):
                            ['shap', 'weight', 'gain', 'cover', 'total_gain', 'total_cover'].
    :param max_depth (int): Maximum tree depth for base learners (default = 6). Must be greater equal to 0.
    :param learning_rate (float): Boosting learning rate (default = 0.3). Must be greater equal to 0.
    :param stopping_method (str): 'num_boost_round' or 'early_stopping_rounds' (default = 'num_boost_round').
    :param stopping_rounds (int): Number of rounds for boosting or early stopping (default = 100).
                                  Must be greater than 0.
    :return: DataFrame. Importance score per gene according to each of the input measures.
                        Higher scores are given to more important (informative) genes.
    """

    try:
        validate_input_gi(data, measures, max_depth, learning_rate, stopping_method, stopping_rounds)

        # Filter singletons (strain types with a single related isolate)
        st_lst = data.iloc[:, -1]
        st, counts = np.unique(st_lst, return_counts=True)
        data = data[[x in st[counts > 1] for x in st_lst]].reset_index(drop=True)
        if data.empty:
            raise ValueError(f"Error: 'data' contains only singletons (strain types with a single related isolate).")

        X, y = data.iloc[:, :-1], data.iloc[:, -1]
        results = get_gene_importance(X, y, measures, max_depth, learning_rate, stopping_method, stopping_rounds)
        return results

    except ValueError as ve:
        print(ve)


def gene_reduction_analysis(data, gene_importance, measure, reduction=0.2, linkage_method='complete',
                            percentiles=[0.5, 1], find_recommended_percentile=False,
                            percentiles_to_check = np.arange(.5, 20.5, 0.5), simulated_samples=0, plot_results=True,
                            n_jobs=mp.cpu_count()):
    """
    This function analyzes how minimizing the number of genes in the MLST scheme impacts strain typing performance.
    At each iteration, a reduced subset of most important genes is selected; and based on the allelic profile composed of these genes,
    isolates are clustered into cluster types (CTs) using a distance-based hierarchical clustering.
    The distance between every pair of isolates is measured by a normalized Hamming distance, which quantifies
    the proportion of the genes which disagree on their allelic assignment. The distance between any two clusters
    is determined by the 'linkage' input parameter.

    To obtain the partition into strains (i.e. the induced CTs), we apply a threshold on the distance between isolates
    belonging to the same cluster-type. we use percentile-based thresholds instead of constant thresholds, i.e. for a
    given percentile of distances' distribution, we dynamically calculate the threshold value for each subset of genes.
    The percentile (or percentiles) can be defined by the user in the `percentiles` input parameter, or alternatively
    being selected by the **<em>find recommended percentile</em>** procedure (depicted below) that searches in the
    search space of `percentiles_to_check` input parameter.

    Typing performance is measure by the Adjusted Rand Index (ARI), which quantifies similarity between the induced
    partition into cluster types (that is based on a subset of genes) and the original partition in cluster types (that is based on
    all genes, and was given as an input) (see: https://link.springer.com/article/10.1007/BF01908075).
    p-value for the ARI is calculated using a Monte Carlo simulation study (see: https://www.sciencedirect.com/science/article/abs/pii/S0950329313000852)
    The analysis results are provided as a DataFrame output and are also plotted by default.

    'find recommended percentile' procedure:
    This procedure uses an heuristic to find a recommended percentile. This is the percentile
    with the best overall predictive performance, which is equivalent to the ARI curve with the highest AUC, and is
    referred as ‘best’. At first, we initialize ‘best’ to the minimal percentile in search space (`percentiles_to_check`).
    Then we compare ‘best’ to the successor percentile in `percentiles_to_check`, referred as ‘next’,
    by computing the "non-absolute" L1 distance between their ARI vectors.
    This distance equals to the sum of the differences between the two vectors when subtracting the 'best' from the 'next'.
    In case the distance is not negative (i.e., 'next' performs better or the same), 'next' is defined as the new 'best'.
    Otherwise, the search is completed and 'best' is selected as the recommended percentile.



    :param data (DataFrame): DataFrame in the shape of (m,n).
                         (n-1) columns of genes, last column (n) must contain the CT (cluster type).
                         Each row (m) represents a profile of a single isolate.
                         Data types should be integers only.
                         Missing values should be represented as 0, no missing values are allowed for the CT (last
                         column).
    :param gene_importance (DataFrame): Importance scores in the format returned by 'gene_importance' function.
    :param measure (str): A single measure according to which gene importance will be defined.
                          Can be either 'shap', 'weight', 'gain', 'cover', 'total_gain' or 'total_cover'.
                          Note that the selected measure must be included in the `gene_importance` input
    :param reduction (numeric): The number (int) or percentage (0<float<1) of least important genes to be removed at
                                each iteration (default = 0.2).
                                The first iteration includes all genes, the second iteration includes all informative
                                genes (importance score > 0), and the subsequent iterations include a reduced
                                subset of genes according to the `reduction` parameter
    :param linkage_method (str): The linkage method to compute the distance between clusters in the hierarchical
                                 clustering algorithm (default = 'complete').
                                 Can be either 'single', 'complete', 'average', 'weighted', 'centroid', 'median' or
                                 'ward'.
                                 for more info: https://docs.scipy.org/doc/scipy/reference/generated/scipy.cluster.hierarchy.linkage.html
    :param percentiles (float or array of floats): The percentile (or percentiles) of distances distribution to be used
                                                   (default = [0.5, 1]).
                                                   Each percentile must be greater than 0 and smaller than 100.
                                                   For a given percentile, we dynamically calculate the threshold value
                                                   for each subset of genes.
                                                   The threshold value refers to the distance between isolates of the
                                                   same CT, which is defined as the proportion of genes that disagree on
                                                   their allelic assignment.
    :param find_recommended_percentile (boolean): if True, ignore parameter 'percentiles' and run a procedure to find a
                                                  recommended threshold out of 'percentiles_to_check' (default = False).
                                                  The outputs will be the ARI and p-value results computed for each
                                                  subset of genes when using the recommended percentile.
    :param percentiles_to_check (array of floats): The percentiles of distances distribution to be evaluated by the
                                                   'find recommended percentile' procedure (default = numpy.arange(.5, 20.5, 0.5)).
                                                   The array must contain at least 2 percentiles; each percentile must
                                                   be greater than 0 and smaller than 100.
    :param simulated_samples (int): The number of samples (partitions) to simulate the computation of the
                                    p-value of the observed ARI (default = 0).
                                    For the significance of the p-value results, it's recommended to use ~1000 samples
                                    or more (see- https://www.sciencedirect.com/science/article/pii/S0950329313000852).
                                    In case 'simulated_samples'=0, simulation won't run and p-value won't be calculated.
    :param plot_results (boolean): if True, plot the ARI and p-value results for each selected percentile
                                   as a function of the number of genes (default = True).
    :param n_jobs (int): The maximum number of concurrently running jobs.
                         (default = min(60, number of CPUs in the system)).
    :return: DataFrame. ARI and p-value (if 'simulated_samples' > 0) computed for each subset of most important genes,
                        and for each selected percentile.
    """
    try:
        n_jobs, percentiles_to_check = validate_input_gra(data, gene_importance, measure, reduction, linkage_method,
                                                          percentiles, percentiles_to_check, simulated_samples, n_jobs)
        # remove non-informative genes
        gi = gene_importance[gene_importance['importance_by_' + measure] > 0]
        num_informative = len(gi)
        print(f"{num_informative} informative genes were found")
        if reduction < 1:
            reduction = int(num_informative*reduction)
        # lst = np.arange(num_informative, 0, -reduction)
        lst = [len(gene_importance)] + list(np.arange(num_informative, 0, -reduction))
        # sort importance according to selected measure
        gene_importance = gene_importance.sort_values(by='importance_by_' + measure, ascending=False).reset_index(drop=True)
        X, CT = data.iloc[:, :-1], data.iloc[:, -1]

        print("Hierarchical clustering")
        try:
            results = Parallel(n_jobs=n_jobs, verbose=5, max_nbytes=None)(
                delayed(hierarchical_clustering)(CT, X, num_of_genes, gene_importance, linkage_method, percentiles,
                                                 find_recommended_percentile, percentiles_to_check, simulated_samples)
                for num_of_genes in lst)
        except Exception as ex:
            print(f"Error - unable to perform parallel computing due to: {ex}")
            print(traceback.format_exc())
            print(f"Running serial computation instead")
            results = []
            for num_of_genes in lst:
                r = hierarchical_clustering(CT, X, num_of_genes, gene_importance, linkage_method, percentiles,
                                            find_recommended_percentile, percentiles_to_check, simulated_samples)
                results = results + [r]

        analysis_res = pd.DataFrame(results)

        if find_recommended_percentile:
            analysis_res = find_percentile(analysis_res, CT, percentiles_to_check, simulated_samples, n_jobs)
        analysis_res = reorder_analysis_res(analysis_res)
        f_analysis_res = analysis_res.loc[:, ~analysis_res.columns.str.startswith("threshold")]

        if plot_results:
            plot_res(f_analysis_res, measure)

        return f_analysis_res

    except ValueError as ve:
        print(ve)









import multiprocessing as mp
import os
from os.path import join
import numpy as np

SEED = 10

# region params gene importance

OBJECTIVE = 'multi:softmax'
EVAL_METRIC = ['merror', 'mlogloss']
# merror: Multiclass classification error rate. It is calculated as #(wrong cases)/#(all cases).

# endregion

# region params hierarchical clustering

DISTANCE_METRIC = 'matching'  # 'hamming'
# Distance is normalized so it ranges between [0, 1]
# (proportion of those vector elements between two n-vectors u and v which disagree.)
TEMP_FOLDER = join(os.getcwd(), 'mlst_temp')

# endregion




import re
import codecs
from setuptools import setup, find_packages
from os.path import join, abspath, dirname


def get_version(*file_paths):
    with codecs.open(join(abspath(dirname(__file__)), *file_paths), 'r') as fp:
        file = fp.read()
    version = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]", file, re.M)
    if version:
        return version.group(1)
    raise RuntimeError("Version string wasn't found.")


with open("README.md", "r") as fh:
    long_description = fh.read()


setup(
    name="minmlst",
    version=get_version("minmlst", "__init__.py"),
    author="Shani Cohen",
    author_email="shani.cohen.33@gmail.com",
    description="Machine-learning based minimal MLST scheme for bacterial strain typing",
    long_description="minMLST is a machine-learning based methodology for identifying a minimal subset of genes " +
                     "that preserves high discrimination among bacterial strains. It combines well known " +
                     "machine-learning algorithms and approaches such as XGBoost, distance-based hierarchical " +
                     "clustering, and SHAP. \nminMLST quantifies the importance level of each gene in an MLST " +
                     "scheme and allows the user to investigate the trade-off between minimizing the number " +
                     "of genes in the scheme vs preserving a high resolution among strains.\n\n " +
                     "See more information in [GitHub](https://github.com/shanicohen33/minMLST).",
    long_description_content_type="text/markdown",
    url="https://github.com/shanicohen33/minMLST",
    packages=find_packages(),
    install_requires=[
        'shap>=0.28.5',
        'xgboost>=0.82',
        'dill>=0.3.0',
        'matplotlib>=3.0.3',
        'IPython>=6.1.0',
        'joblib>=0.13.2',
        'scikit-learn==0.20.1'
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        'Programming Language :: Python :: 3.6',
        "License :: OSI Approved :: MIT License",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: POSIX :: Linux"
    ],
)




import minmlst.config as c
import numpy as np
import xgboost as xgb
import pandas as pd
import shap
import traceback
from collections import OrderedDict
from sklearn.preprocessing import LabelEncoder
from sklearn.model_selection import StratifiedKFold


def calc_shap_values(bst, X, y):
    print("  mean_shap_values")
    explainer = shap.TreeExplainer(bst)
    shap_values = explainer.shap_values(X=X, y=y)
    del explainer
    mean_shap_values = np.sum(np.mean(np.abs(shap_values), axis=0), axis=0)
    del shap_values
    gene_importance = pd.DataFrame({'gene': list(X.columns.values),
                                    'importance_by_shap': list(mean_shap_values)})
    return gene_importance


def calc_importance(bst, measure):
    print(f"  {measure}")
    scores = bst.get_score(importance_type=measure)
    gene_importance = pd.DataFrame({'gene': list(scores.keys()),
                                    'importance_by_' + measure: list(scores.values())})
    return gene_importance


def get_gene_importance(X, y, measures, max_depth, learning_rate, stopping_method, stopping_rounds):
    # encode y labels
    lbl = LabelEncoder()
    lbl.fit(y)
    y_enc = lbl.transform(y)
    n_classes = len(np.unique(y_enc))

    # split into train and test (stratified)
    skf = StratifiedKFold(n_splits=2, shuffle=False)
    for train_index, test_index in skf.split(X=X, y=y_enc):
        if len(train_index) < len(test_index):
            tmp = train_index
            train_index = test_index
            test_index = tmp

        X_train = X.iloc[list(train_index), :]
        y_train = y_enc[train_index]
        X_test = X.iloc[list(test_index), :]
        y_test = y_enc[test_index]
        break
    dtrain = xgb.DMatrix(X_train, label=y_train)
    dtest = xgb.DMatrix(X_test, label=y_test)

    # train an xgboost model
    params = {
        'seed': c.SEED,
        'objective': c.OBJECTIVE,
        'num_class': n_classes,
        'max_depth': max_depth,
        'eta': learning_rate,
        'verbose_eval': True,
        "silent": 1,
        'eval_metric': c.EVAL_METRIC
    }
    watchlist = [(dtrain, 'train'), (dtest, 'test')]
    evals_result = {}

    print("Training a model")
    if stopping_method == 'early_stopping_rounds':
        bst = xgb.train(
            early_stopping_rounds=stopping_rounds,
            params=params,
            dtrain=dtrain,
            evals=watchlist,
            verbose_eval=10,
            evals_result=evals_result
        )
    else:
        bst = xgb.train(
            num_boost_round=stopping_rounds,
            params=params,
            dtrain=dtrain,
            evals=watchlist,
            verbose_eval=10,
            evals_result=evals_result
        )

    print(f"Calculating gene importance by:")
    scores_df = pd.DataFrame({'gene': list(X.columns.values)})
    for m in list(OrderedDict.fromkeys(measures)):
        if m == "shap":
            try:
                scores = calc_shap_values(bst=bst, X=X, y=y_enc)
            except Exception as ex:
                print(f"Error - unable compute SHAP values due to: {ex}")
                print(traceback.format_exc())
                scores = pd.DataFrame()
        else:
            scores = calc_importance(bst=bst, measure=m)
        if not scores.empty:
            scores_df = pd.merge(scores_df, scores, how='left', on='gene')
    scores_df = scores_df.fillna(0).sort_values(by=scores_df.columns.values[1], ascending=False).reset_index(drop=True)

    return scores_df




import os


def get_dir_files(dir_path):
    files = [file for file in os.listdir(dir_path) if os.path.isfile(os.path.join(dir_path, file))]
    return sorted(files)


def create_dir_if_not_exists(dir_path):
    if not os.path.exists(dir_path):
        os.makedirs(dir_path)


def remove_dir(dir_path):
    for file in os.listdir(dir_path):
        file_path = os.path.join(dir_path, file)
        if os.path.isfile(file_path):
            os.unlink(file_path)
    os.rmdir(dir_path)
